@@ -6,9 +6,11 @@ extern "C" {
 #endif
 
 #define COMMAND_LINE_SIZE	1024
-#define PAGE_OFFSET		0xc0000000
+#define DEFAULT_PAGE_OFFSET		(0xc0000000)
+#define KVBASE_MASK	(0x1ffffff)
 #define CRASH_MAX_MEMORY_RANGES	32
 
+
 extern struct memory_ranges usablemem_rgns;
 
 struct kexec_info;
